//      Win API     .  .

unit MakeInstance;

interface

uses
  Windows, Messages;

function MakeInstanceStdCall(Obj: TObject; MethodAddr: Pointer): Pointer;
procedure FreeInstanceStdCall(Inst: Pointer);

type
  TMsgEvent =
    procedure(Sender: TObject; var Message: TMessage; var Handled: boolean)
                                                                   of Object;

  TWndObject = class
  private
    FHandle: HWND;
    FOnMessage: TMsgEvent;
    function ObjWindowProc(Wnd: HWND; AMsg: UINT;
                           wPrm: WPARAM; lPrm: LPARAM): LRESULT; stdcall;
    function GetWindowName: string;
    procedure SetWindowName(const Value: string);
  protected
    procedure GetClassParams(var WndClass: TWndClass); virtual;
    procedure GetStyles(var Style, ExStyle: DWORD); virtual;
    procedure DoMessage(var Message: TMessage); virtual;
  public
    constructor Create(const AName: string = ''); virtual;
    destructor Destroy; override;
    property Handle: HWND read FHandle;
    property OnMessage: TMsgEvent read FOnMessage write FOnMessage;
    {      OnMessage}
    property WindowName: string read GetWindowName write SetWindowName;
  end;

implementation

var
  FCS: TRTLCriticalSection;

{ TWndObject }

const
  WndClassTemplate: TWndClass = (
    style: 0;
    lpfnWndProc: @DefWindowProc;
    cbClsExtra: 0;
    cbWndExtra: 0;
    hInstance: 0;
    hIcon: 0;
    hCursor: 0;
    hbrBackground: 0;
    lpszMenuName: nil;
    lpszClassName: 'TSNUtilWindow');

constructor TWndObject.Create(const AName: string = '');
var
  WndClass, TempClass: TWndClass;
  Style, ExStyle: DWORD;
  ClassRegistered: Boolean;
begin
  GetClassParams(WndClass);
  GetStyles(Style, ExStyle);
  if IsMultiThread then EnterCriticalSection(FCS);
  try
    ClassRegistered := GetClassInfo(HInstance, WndClass.lpszClassName,
      TempClass);
    if ClassRegistered then UnregisterClass(WndClass.lpszClassName, HInstance);
    RegisterClass(WndClass);
    FHandle := CreateWindowEx(ExStyle, WndClass.lpszClassName,
                        PChar(AName), Style, 0, 0, 0, 0, 0, 0, HInstance, nil);
    if WndClass.lpfnWndProc = @DefWindowProc then
      SetWindowLong(FHandle, GWL_WNDPROC,
        Longint(MakeInstanceStdCall(Self, @TWndObject.ObjWindowProc)));
  finally
    if IsMultiThread then LeaveCriticalSection(FCS);
  end;
end;

destructor TWndObject.Destroy;
var
  Instance: Pointer;
begin
  Instance := Pointer(GetWindowLong(FHandle, GWL_WNDPROC));
  DestroyWindow(FHandle);
  if Instance <> @DefWindowProc then FreeInstanceStdCall(Instance);
  inherited;
end;

procedure TWndObject.DoMessage(var Message: TMessage);
var
  Handled: boolean;
begin
  Handled:=false;
  if Assigned(FOnMessage) then FOnMessage(Self, Message, Handled);
  if not Handled then with Message do
    Result:=DefWindowProc(FHandle, Msg, WParam, LParam);
end;

procedure TWndObject.GetClassParams(var WndClass: TWndClass);
begin
  WndClass:=WndClassTemplate;
  WndClass.hInstance:=HInstance;
end;

procedure TWndObject.GetStyles(var Style, ExStyle: DWORD);
begin
  Style:=WS_POPUP;
  ExStyle:=WS_EX_TOOLWINDOW;
end;

function TWndObject.GetWindowName: string;
var
  n: DWORD;
begin
  n:=GetWindowTextLength(FHandle);
  if n = 0 then Exit;
  SetLength(Result, n);
  GetWindowText(FHandle, @Result[1], n + 1);
end;

function TWndObject.ObjWindowProc(Wnd: HWND; AMsg: UINT; wPrm: WPARAM;
  lPrm: LPARAM): LRESULT;
var
  Message: TMessage;
begin
  with Message do
  begin
    Msg:=AMsg;
    WParam:=wPrm;
    LParam:=lPrm;
    Result:=0;
  end;
  DoMessage(Message);
  Result:=Message.Result;
end;

procedure TWndObject.SetWindowName(const Value: string);
begin
  SetWindowText(FHandle, PChar(Value));
end;

{ /TWndObject }

{
   -   TInstanceBlock  16 ,   
     -    .
}

type
  PInstanceBlock = ^TInstanceBlock;
  TInstanceBlock = packed record
    Code:  array[0..31 - SizeOf(PInstanceBlock)] of Byte;
    Next:  PInstanceBlock;
  end;

var
  BlockList: PInstanceBlock = nil;
  Region: Pointer = nil;
  NextPage: pointer = nil;
  BlockCount: integer = 0;

const
  Page_Size = 4096;
  BlocksOnPage = Page_Size div SizeOf(TInstanceBlock);

type
  TBlockList = array[0..BlocksOnPage - 1] of TInstanceBlock;
  PBlockList = ^TBlockList;

function GetAllocGran: DWORD;
var
  Info: TSystemInfo;
begin
  GetSystemInfo(Info);
  Result:=Info.dwAllocationGranularity;
end;

procedure FreeInstanceStdCall(Inst: Pointer);
begin
  if Inst = nil then Exit;
  if IsMultiThread then EnterCriticalSection(FCS);
  PInstanceBlock(Inst).Next:=BlockList;
  BlockList:=Inst;
  Dec(BlockCount);
  if BlockCount = 0 then
  begin
    VirtualFree(Region, 0, MEM_RELEASE);
    Region:=nil;
    NextPage:=nil;
    BlockList:=nil;
  end;
  if IsMultiThread then LeaveCriticalSection(FCS);
end;

function MakeInstanceStdCall(Obj: TObject; MethodAddr: Pointer): Pointer;
const
  CodeBlock: array [0..12] of Byte =
    (
     $58,             // POP EAX
     $68,             // PUSH DWORD_Const <= Obj
      0, 0, 0, 0,
     $50,             // PUSH EAX
     $68,             // PUSH DWORD_Const <= MethodAddr
      0, 0, 0, 0,
     $C3              // RET
     );

type
  PCodeBlock = ^TCodeBlock;
  TCodeBlock = packed record
    Code1:     Word;
    Self:      TObject;
    Code2:     Word;
    MtdAddr:   Pointer;
    Code3:     Byte;
  end;

var
  n: integer;
begin
  if IsMultiThread then EnterCriticalSection(FCS);

  if BlockList = nil then
  begin
    if Region = nil then
    begin
      Region:=VirtualAlloc(nil, GetAllocGran, MEM_RESERVE or MEM_TOP_DOWN,
                                                     PAGE_EXECUTE_READWRITE);
      NextPage:=Region;
    end;
    BlockList:=VirtualAlloc(NextPage, Page_Size, MEM_COMMIT,
                                       PAGE_EXECUTE_READWRITE);
    Inc(Cardinal(NextPage), Page_Size);
    for n:=0 to BlocksOnPage - 2 do
      PBlockList(BlockList)[n].Next:=@PBlockList(BlockList)[n + 1];
    PBlockList(BlockList)[BlocksOnPage - 1].Next:=nil;
  end;
  Result:=BlockList;
  BlockList:=BlockList.Next;
  Inc(BlockCount);

  if IsMultiThread then LeaveCriticalSection(FCS);

  Move(CodeBlock, Result^, SizeOf(CodeBlock));
  with PCodeBlock(Result)^ do
  begin
    Self:=Obj;
    MtdAddr:=MethodAddr;
  end;
end;

procedure InitProc;
begin
  InitializeCriticalSection(FCS);
end;

procedure FinalProc;
begin
  DeleteCriticalSection(FCS);
end;

initialization
  InitProc;

finalization
  FinalProc;

end.
